/*
 *
 * motor.h
 *	Control Motor
 *
 *
 * Copyright 2008-2009 
 * Uchibori Laboratory, Numazu National College of Technology
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */


#ifndef MG3_MOTOR_H
#define MG3_MOTOR_H

#include <sys/types.h>
#include <stdint.h>
#include <time.h>
#include <mg3/motor_config.h>





//
// Low level handler
//

/*
 * Protocol
 * # Ping
 * Req: 0x00
 * Res: 0x80
 *
 * # 制御パラメータの設定
 * Req: 0x01, 目標値, Kp, Ki
 * Res: 0x81
 *  目標値: -128〜127(パルス/100ms)
 *  Kp: 0〜255
 *  Ki: 0〜255
 *
 * # モータの状態取得
 * Req: 0x02
 * Res: 0x82, 速度, PWMデューティ(2Byte), 累積カウント(4Byte)
 *  速度: -128〜127[パルス/100ms]
 *  PWM duty: 0〜255 (duty比: 0〜100%)(要検討)
 *  累積カウント: -2^31〜2^31-1[パルス]
 *
 * # スタート
 * Req: 0x04
 * Res: 0x84
 *
 * # ストップ
 * Req: 0x08
 * Res: 0x88
 *
 * # 制御パラメータの取得
 * Req: 0x10
 * Res: 0x90, 目標値, Kp, Ki
 *
 * # Unknown
 * Req: 0x??
 * Res: 0x40 | Req
 */

// Contorol packet ID
#define MG3_MOTOR_PKT_PING  	0x00
#define MG3_MOTOR_PKT_SETCTL	0x01
#define MG3_MOTOR_PKT_GETST 	0x02
#define MG3_MOTOR_PKT_START 	0x04
#define MG3_MOTOR_PKT_STOP  	0x08
#define MG3_MOTOR_PKT_GETCTL	0x10
#define MG3_MOTOR_PKT_ACK   	0x80


ssize_t _mg3_motor_recv_data (int fd, void *buf, size_t size, long timeout);
ssize_t _mg3_motor_send_data (int fd, void *buf, size_t size, long timeout);
int motor_get_sci_error (int fd);



int motor_send_control_pkt (int fd, uint8_t speed, uint8_t kp, uint8_t ki);
//int motor_recv_control_pkt (int fd, uint8_t *speed, uint8_t *kp, uint8_t *ki);
int motor_recv_status_pkt (int fd, uint8_t *speed, uint16_t *pwm, uint32_t *count);











typedef struct MG3Motor MOTOR;



struct MotorControl {
	double r;
	double kp;
	double ki;
};



struct MotorState {
	struct timespec ctrled_time;
	double speed;
	double total_round;
};



MOTOR *mg3_motor_open_by_fd (int fd);
int mg3_motor_get_ctrl (const MOTOR *motor, struct MotorControl *ctrl);
int mg3_motor_set_ctrl (MOTOR *motor, const struct MotorControl *ctrl);
int mg3_motor_get_state (const MOTOR *motor, struct MotorState *state);



//
// handle display infomation
//

#define MG3_MESSAGE_NONE  0
#define MG3_MESSAGE_ERROR 1 
#define MG3_MESSAGE_INFO  2
#define MG3_MESSAGE_DEBUG 3


int mg3_motor_set_message_level (int level);


#endif // MG3_MOTOR_H

