/*
 *
 * comm.c
 *	シリアル通信
 *
 *
 * Copyright 2008-2009 
 * Uchibori Laboratory, Numazu National College of Technology
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */


#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/select.h>
#include "mctrl.h"




//
// 制限時間なしでデータを読み込む
//
static ssize_t recv_data (int fd, void *buf, const size_t size)
{
	ssize_t result;           // 関数の返り値
	ssize_t read_size = 0;    // 読み込めたデータ数
	unsigned char *p = buf;   // データを書き込むバッファの位置


	while ( read_size < (ssize_t)size ) {
		// シリアルポートからデータを読み込む．
		result = read(fd, p, size - read_size);
		if ( result >= 0 ) {
			// 読み込みに(一部)成功した．
			read_size += result;
			p += result;
		}
		else {
			//
			// 何かエラーが起きた．
			//

			if ( errno == EINTR ) {
				// シグナルに割り込まれたので，もう一度やり直す．
				continue;
			}
			
			if ( errno == EAGAIN ) {
				// 非ブロックI/Oに設定されていてデータが全くなかった．
				// ブロックI/Oに設定して，やり直す．
				int fflag = fcntl(fd, F_GETFL);
				if ( fflag == -1 ) {
					return -1;
				}
				fflag &= ~O_NONBLOCK;
				result = fcntl(fd, F_SETFL, fflag);
				if ( result < 0 ) {
					return -1;
				}

				continue;
			}

			// 想定してないエラーなので
			// 対処はユーザに任せる．
			return read_size > 0 ? read_size : -1;
		} // end of if ( result < 0 )
	} // end of while ( read_size < size )


	assert((size_t)read_size == size);
	return read_size;
}



//
// 制限時間付きでデータを読み込む
//
static ssize_t recv_data_timeout (int fd, void *buf, const size_t size, long wait)
{
	struct timeval wait_time; // 最大待ち時間
	fd_set rfds;              // どのシリアルポートも待つのか指定するためのもの
	ssize_t result;           // 関数の返り値
	ssize_t read_size = 0;    // 読み込めたデータ数
	int old_fflag, new_fflag; // 非ブロックI/Oに切り替えるためのfcntl()用．
	char *p = buf;            // データを書き込むバッファの位置


	// 非ブロックI/Oに切り替える．
	old_fflag = fcntl(fd, F_GETFL);
	new_fflag = old_fflag | O_NONBLOCK;
	if ( old_fflag == -1 ) {
		return -1;
	}
	if ( (old_fflag & O_NONBLOCK) == 0 ) {
		result = fcntl(fd, F_SETFL, new_fflag);
		if ( result < 0 ) {
			return -1;
		}
	}

	// ナノ秒からマイクロ秒へ切り上げる．
	wait += 999;
	wait /= 1000;

	wait_time.tv_sec = wait / 1000000L;
	wait_time.tv_usec = (suseconds_t)(wait - wait_time.tv_sec);

	FD_ZERO(&rfds);
	FD_SET(fd, &rfds);
	
	
	while ( read_size < (ssize_t)size ) {
		// データが読める状態になるのを待つ．
		result = select(fd + 1, &rfds, NULL, NULL, &wait_time);
		if ( result == 0 ) {
			// 時間切れなので，処理を中断する．
			if ( read_size > 0 ) {
				return read_size;
			}
			else {
				errno = ETIMEDOUT;
				return -1;
			}
		}

		if ( result < 0 ) {
			if ( errno == EINTR ) {
				// シグナルに割り込まれた．
				// Linuxのマニュアルによると，シグナルに割り込まれても，
				// wait_timeの値を変更してくれるらしいので，
				// そのまま使用して，select()を続ける．
				continue;
			}
			else {
				return -1;
			}
		}

		// fdしか待っていないので，
		// それ以外のものが読み込み可能になることはありえないが，一応念のため．
		if ( !FD_ISSET(fd, &rfds) ) {
			continue;
		}

		// ここまで処理が進んだということは，
		// 少なくとも1文字は読めるはず．

		// シリアルポートからデータを読み込む．
		result = read(fd, p, size - read_size);
		if ( result >= 0 ) {
			// 読み込みに(一部)成功した．
			read_size += result;
			p += result;
		}
		else {
			// 何かエラーが起きた．

			if ( errno == EINTR ) {
				// シグナルに割り込まれたので，もう一度やり直す．
				continue;
			}

			// 想定してないエラーなので
			// 対処はユーザに任せる．
			return read_size > 0 ? read_size : -1;
		}
	} // end of while ( read_size < size )


	// I/Oフラグを元にもどす．
	if ( old_fflag != new_fflag ) {
		// エラーが起こっても無視．
		fcntl(fd, F_SETFL, old_fflag);
	}


	assert((size_t)read_size == size);
	return read_size;
	return 0;
}



ssize_t _mg3_motor_recv_data (int fd, void *buf, size_t size, long wait)
{
	if ( (ssize_t)size < 0 ) {
		errno = EINVAL;
		return -1;
	}

	if ( wait < 0 ) {
		// ブロックI/Oによる読み込み
		return recv_data(fd, buf, size);
	}
	else {
		// 非ブロックI/Oによる読み込み
		return recv_data_timeout(fd, buf, size, wait);
	}
}






//
// 制限時間なしでデータを書き込む
//
static ssize_t send_data (int fd, const void *buf, const size_t size)
{
	ssize_t result;           // 関数の返り値
	ssize_t written_size = 0; // 書き込めたデータ数
	const char *p = buf;      // データを読み込むバッファの位置


	while ( written_size < (ssize_t)size ) {
		// シリアルポートにデータを書き込む．
		result = write(fd, p, size - written_size);
		if ( result >= 0 ) {
			written_size += result;
			p += result;
		}
		else {
			//
			// 何かエラーが起きた．
			//

			if ( errno == EINTR ) {
				// シグナルに割り込まれたので，もう一度やり直す．
				continue;
			}
			
			if ( errno == EAGAIN ) {
				// 非ブロックI/Oに設定されていてデータが全く書き込めなかった．
				// ブロックI/Oに設定して，やり直す．
				int fflag = fcntl(fd, F_GETFL);
				if ( fflag == -1 ) {
					return -1;
				}
				fflag &= ~O_NONBLOCK;
				result = fcntl(fd, F_SETFL, fflag);
				if ( result < 0 ) {
					return -1;
				}

				continue;
			}

			// 想定してないエラーが起こった．
			// 対処はユーザに任せる．
			return written_size > 0 ? written_size : -1;
		} // end of if ( result < 0 )
	} // end of while ( read_size < size )


	assert((size_t)written_size == size);
	return written_size;
}



//
//制限時間付きでデータを書き込む
//
static ssize_t send_data_timeout (int fd, void *buf, const size_t size, long wait)
{
	struct timeval wait_time; // 最大待ち時間
	fd_set wfds;              // どのシリアルポートも待つのか指定するためのもの
	ssize_t result;           // 関数の返り値
	ssize_t written_size = 0; // 書き込めたデータ数
	int fflag;                // 非ブロックI/Oに切り替えるためのfcntl()用．
	unsigned char *p = buf;   // データを読み込むバッファの位置


	// 非ブロックI/Oに切り替える．
	fflag = fcntl(fd, F_GETFL);
	if ( fflag == -1 ) {
		return -1;
	}
	if ( (fflag & O_NONBLOCK) == 0 ) {
		fflag |= O_NONBLOCK;
		result = fcntl(fd, F_SETFL, fflag);
		if ( result < 0 ) {
			return -1;
		}
	}

	// ナノ秒からマイクロ秒へ切り上げる．
	wait += 999;
	wait /= 1000;

	wait_time.tv_sec = wait / 1000000L;
	wait_time.tv_usec = (suseconds_t)(wait - wait_time.tv_sec);

	FD_ZERO(&wfds);
	FD_SET(fd, &wfds);
	
	
	while ( written_size < (ssize_t)size ) {
		// データが書ける状態になるのを待つ．
		result = select(fd + 1, NULL, &wfds, NULL, &wait_time);
		if ( result == 0 ) {
			// 時間切れなので，処理を中断する．
			if ( written_size > 0 ) {
				return written_size;
			}
			else {
				errno = ETIMEDOUT;
				return -1;
			}
		}

		if ( result < 0 ) {
			if ( errno == EINTR ) {
				// シグナルに割り込まれた．
				// Linuxのマニュアルによると，シグナルに割り込まれても，
				// wait_timeの値を変更してくれるらしいので，
				// そのまま使用して，select()を続ける．
				continue;
			}
			else {
				return -1;
			}
		}

		// fdしか待っていないので，
		// それ以外のものが書き込み可能になることはありえないが，一応念のため．
		if ( !FD_ISSET(fd, &wfds) ) {
			continue;
		}

		// ここまで処理が進んだということは，
		// 少なくとも1文字は書けるはず．

		// シリアルポートにデータを書き込む．
		result = write(fd, p, size - written_size);
		if ( result >= 0 ) {
			// 書き込みに(一部)成功した．
			written_size += result;
			p += result;
		}
		else {
			// 何かしらのエラーが起きた．

			if ( errno == EINTR ) {
				// シグナルに割り込まれたので，もう一度やり直す．
				continue;
			}

			// 想定してないエラーが起こった．
			// 対処はユーザに任せる．
			return written_size > 0 ? written_size : -1;
		}
	} // end of while ( read_size < size )


	assert((size_t)written_size == size);
	return written_size;
	return 0;
}



ssize_t _mg3_motor_send_data (int fd, void *buf, size_t size, long wait)
{
	if ( (ssize_t)size < 0 ) {
		errno = EINVAL;
		return -1;
	}

	if ( wait < 0 ) {
		// ブロックI/Oによる読み込み
		return send_data(fd, buf, size);
	}
	else {
		// 非ブロックI/Oによる読み込み．
		return send_data_timeout(fd, buf, size, wait);
	}
}



