/*
 *
 * open.c
 *	Create data and thread to control motor
 *
 *
 * Copyright 2008-2009 
 * Uchibori Laboratory, Numazu National College of Technology
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */


#include <assert.h>
#include <errno.h>
#include <pthread.h>
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <mg3/motor.h>
#include "mctrl.h"
#include "message.h"


typedef void *(*thread_start)(void *);


static void *control_thread (MOTOR *arg)
{
	int i;
	int err;
	struct timespec wake_time;
	MOTOR *motor;


	motor = arg;
	assert(motor->ctrl_thread == pthread_self());


	err = clock_gettime(CLOCK_ID, &wake_time);
	if ( err != 0 ) {
		return (void *)err;
	}


	if ( mg3_motor_message_need_to_display(MG3_MESSAGE_INFO) ) {
		int policy;
		struct sched_param schd_prm;

		err = pthread_getschedparam(pthread_self(), &policy, &schd_prm);
		if ( err == 0 ) {
			const char *policy_string;
			switch (policy) {
			case SCHED_FIFO:
				policy_string = "SCHED_FIFO";
				break;
			case SCHED_RR:
				policy_string = "SCHED_RR";
				break;
			case SCHED_OTHER:
				policy_string = "SCHED_OTHER";
				break;
			default:
				policy_string = "unknown";
				break;
			}
			fprintf(stderr, "Thread for motor control is ready.\n");
			fprintf(stderr, "\tID       : %lu\n", (unsigned long)pthread_self());
			fprintf(stderr, "\tPolicy   : %s\n", policy_string);
			fprintf(stderr, "\tPriority : %d\n\n", schd_prm.sched_priority); 
		}
	}

	for ( i = 0; i < 100; i++ ) {
		pthread_testcancel();

		wake_time.tv_nsec += CYCLE;
		while ( wake_time.tv_nsec >= 1000000000 ) {
			wake_time.tv_sec += 1;
			wake_time.tv_nsec -= 1000000000;
		}

		do {
			err = clock_nanosleep(CLOCK_ID, TIMER_ABSTIME, &wake_time, NULL);
		} while ( err == -1 && errno == EINTR );

		if ( err != 0 ) {
			perror("nano_sleep");
			return (void *)err;
		}

		//
		// control
		//

		// 1秒ごとにビープ
		if ( i % 20 == 0 ) {
			fprintf(stderr, "\a*");
		}
	}
	printf("\n");


	return NULL;
}



#define exit_open(motor, err) do { \
	free(motor); \
	errno = err; \
	return NULL; \
} while (0)

MOTOR *mg3_motor_open_by_fd (int fd)
{
	int err;
	int policy;
	MOTOR *motor;
	pthread_mutexattr_t mutex_attr;
	pthread_attr_t thread_attr;
	struct sched_param schd_prm;


	// 制御用構造体用のメモリを確保する．
	motor = malloc(sizeof(MOTOR));
	if ( motor == NULL ) {
		return NULL;
	}
	memset(motor, 0, sizeof(MOTOR));


	// 通信用ファイルディスクリプタをセットする．
	motor->comm_fd = fd;


	//
	// 同期用mutexを初期化する．
	//
	err = pthread_mutexattr_init(&mutex_attr);
	if ( err != 0 ) {
		exit_open(motor, err);
	}
	// TODO pthread_mutexattr_settype(&mutex_attr, PTHREAD_MUTEX_RECURSIVE) ?
	err = pthread_mutex_init(&motor->mutex, &mutex_attr);
	if ( err != 0 ) {
		exit_open(motor, err);
	}
	// 優先順位の逆転を防ぐため，優先度継承プロトコルを用いる．
	err = pthread_mutexattr_setprotocol(&mutex_attr, PTHREAD_PRIO_INHERIT);
	if ( err != 0 ) {
		exit_open(motor, err);
	}


	//
	// 制御スレッドを生成する．
	//
	err = pthread_attr_init(&thread_attr);
	if ( err != 0 ) {
		exit_open(motor, err);
	}

	err = pthread_attr_setinheritsched(&thread_attr, PTHREAD_EXPLICIT_SCHED);
	if ( err != 0 ) {
		exit_open(motor, err);
	}

	err = pthread_attr_setschedpolicy(&thread_attr, MOTOR_SCHED_POLICY);
	if ( err != 0 ) {
		exit_open(motor, err);
	}

	err = pthread_getschedparam(pthread_self(), &policy, &schd_prm);
	if ( err != 0 ) {
		exit_open(motor, err);
	}

	schd_prm.sched_priority = MOTOR_SCHED_PRIORITY;
	err = pthread_attr_setschedparam(&thread_attr, &schd_prm);
	if ( err != 0 ) {
		exit_open(motor, err);
	}


	err = pthread_create(&motor->ctrl_thread, &thread_attr,
		(thread_start)control_thread, motor);
	if ( err != 0 ) {
		exit_open(motor, err);
	}



	pthread_join(motor->ctrl_thread, NULL);


	return motor;
}

#undef open_error


