/*
 *  fifo.c
 *  
 *	Copyright 2008-2009 
 *	Uchibori Laboratory, Numazu National College of Technology
 *	All rights reserved.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, version 2 of the License.
 *
 *	Original source is kfifo.[h,c] in linux kernel
 *	written by Stelian Pop <stelian@popies.net>
 *
 */


#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/err.h>
#include "fifo.h"





int mg3_fifo_init (struct mg3_fifo *fifo, size_t buf_size, gfp_t gfp_mask, spinlock_t *lock)
{
	unsigned char *buf;

	if ( buf_size & (buf_size - 1) ) {
		BUG_ON(buf_size > 0x80000000);
		buf_size = roundup_pow_of_two(buf_size);
	}

	buf = kmalloc(buf_size, gfp_mask);
	if ( buf == NULL ) {
		return -ENOMEM;
	}

	fifo->buffer = buf;
	fifo->size = buf_size;
	fifo->rp = fifo->wp = 0;
	fifo->lock = lock;


	return 0;
}



void mg3_fifo_destroy (struct mg3_fifo *fifo)
{
	if ( fifo->buffer != NULL ) {
		kfree(fifo->buffer);
		fifo->buffer = NULL;
	}
}



// FIFOからcに1文字読み込む．
// 読み込みに成功したら，0を返す．
int mg3_fifo_getc (struct mg3_fifo *fifo, unsigned char *c)
{
	unsigned long flags;
	int ret;

	spin_lock_irqsave(fifo->lock, flags);

	ret = mg3_fifo_getc_unlocked(fifo, c);
	if ( fifo->wp == fifo->rp ) {
		mg3_fifo_reset_unlocked(fifo);
	}

	spin_unlock_irqrestore(fifo->lock, flags);

	return ret;
}




// FIFOから読み込み位置をシークさせずにcに読み込む．
// 読み込みに成功したら，0を返す.
int mg3_fifo_peek (struct mg3_fifo *fifo, unsigned char *c)
{
	unsigned long flags;
	int ret;

	spin_lock_irqsave(fifo->lock, flags);

	ret = mg3_fifo_peek_unlocked(fifo, c);

	spin_unlock_irqrestore(fifo->lock, flags);

	return ret;
}



// FIFOにcを書き込む．
// 書き込みに成功したら，0を返す.
int mg3_fifo_putc (struct mg3_fifo *fifo, unsigned char c)
{
	unsigned long flags;
	int ret;

	spin_lock_irqsave(fifo->lock, flags);

	ret = mg3_fifo_putc_unlocked(fifo, c);

	spin_unlock_irqrestore(fifo->lock, flags);

	return ret;
}



// FIFO内のデータを破棄する．
void mg3_fifo_reset (struct mg3_fifo *fifo)
{
	unsigned long flags;

	spin_lock_irqsave(fifo->lock, flags);

	mg3_fifo_reset_unlocked(fifo);

	spin_unlock_irqrestore(fifo->lock, flags);
}



// 読み込み位置を1文字分シークする．
// シークに成功したら，0を返す.
int mg3_fifo_seek (struct mg3_fifo *fifo)
{
	unsigned long flags;
	int ret;

	spin_lock_irqsave(fifo->lock, flags);

	ret = mg3_fifo_seek_unlocked(fifo);
	if ( fifo->wp == fifo->rp ) {
		mg3_fifo_reset_unlocked(fifo);
	}

	spin_unlock_irqrestore(fifo->lock, flags);

	return ret;
}



// FIFO内にあるデータ数を返す．
size_t mg3_fifo_size (struct mg3_fifo *fifo)
{
	unsigned long flags;
	size_t ret;

	spin_lock_irqsave(fifo->lock, flags);

	ret = mg3_fifo_size_unlocked(fifo);

	spin_unlock_irqrestore(fifo->lock, flags);

	return ret;
}





// ロック無しでFIFOからcに読み込む．
// 読み込みに成功したら，0を返す.
int mg3_fifo_getc_unlocked (struct mg3_fifo *fifo, unsigned char *c)
{
	if ( fifo->rp == fifo->wp ) {
		return -EAGAIN;
	}
	else {
		*c = fifo->buffer[fifo->rp & (fifo->size - 1)] ;
		++fifo->rp;
		return 0;
	}
}



// ロック無しでFIFOから読み込み位置をシークせずにcに読み込む．
// 読み込みに成功したら，0を返す.
int mg3_fifo_peek_unlocked (struct mg3_fifo *fifo, unsigned char *c)
{
	if ( fifo->rp == fifo->wp ) {
		return -EAGAIN;
	}
	else {
		*c = fifo->buffer[fifo->rp & (fifo->size - 1)] ;
		return 0;
	}
}



// ロック無しでFIFOにcを書き込む．
// 書き込みに成功したら，0を返す.
int mg3_fifo_putc_unlocked (struct mg3_fifo *fifo, unsigned char c)
{
	if ( mg3_fifo_size_unlocked(fifo) == fifo->size ) {
		// バッファがいっぱいなので これ以上書き込めない．
		return -EAGAIN;
	}
	else {
		fifo->buffer[fifo->wp & (fifo->size - 1)] = c;
		++fifo->wp;
		return 0;
	}
}



// ロック無しで読み込み位置を1文字分シークする．
// シークに成功したら，0を返す.
int mg3_fifo_seek_unlocked (struct mg3_fifo *fifo)
{
	if ( fifo->rp == fifo->wp ) {
		return -EAGAIN;
	}
	else {
		++fifo->rp;
		return 0;
	}
}



