/*
 *  fifo.h
 *  
 *	Copyright 2008-2009 
 *	Uchibori Laboratory, Numazu National College of Technology
 *	All rights reserved.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, version 2 of the License.
 *
 *	Original source is kfifo.[h,c] in linux kernel
 *	written by Stelian Pop <stelian@popies.net>
 *
 */


//
// linux kernelのkfifoを改造したもの．
// １文字ずつの入出力に特化したリングFIFO．
// seekさせずに読む，peekも実装した．
//



#ifndef _MG3_RBUF_H
#define _MG3_RBUF_H


#ifdef __KERNEL__

#include <linux/kernel.h>
#include <linux/spinlock.h>


struct mg3_fifo {
	unsigned char *buffer; // データを保持するFIFO．２のべき乗の大きさに切り上げられる．
	size_t size;           // FIFOのサイズ
	volatile size_t wp;    // データはbuffrt[wp % size]に書かれる．
	volatile size_t rp;	   // データはbuffrt[rp % size]にから読まれる．
	spinlock_t *lock;	   // 排他制御用のスピンロック
};


int mg3_fifo_init (struct mg3_fifo *fifo, size_t buf_size, gfp_t gfp_mask, spinlock_t *lock);
void mg3_fifo_destroy (struct mg3_fifo *fifo);



// FIFOからデータを読み込む．
int mg3_fifo_getc (struct mg3_fifo *fifo, unsigned char *c);

// 読み込み位置をシークさせずに読み込む．
int mg3_fifo_peek (struct mg3_fifo *fifo, unsigned char *c);

// FIFOにデータを書き込む．
int mg3_fifo_putc (struct mg3_fifo *fifo, unsigned char c);

// 読み込み位置を1文字分シークする．
int mg3_fifo_seek (struct mg3_fifo *fifo);

// FIFO内にあるデータ数を調べる．
size_t mg3_fifo_size (struct mg3_fifo *fifo);

// FIFO内のデータを破棄する．
void mg3_fifo_reset (struct mg3_fifo *fifo);

// FIFOに書き込める最大データ数を調べる．
static size_t mg3_fifo_capacity (const struct mg3_fifo *fifo);



// without lock
int mg3_fifo_getc_unlocked (struct mg3_fifo *fifo, unsigned char *c);
int mg3_fifo_peek_unlocked (struct mg3_fifo *fifo, unsigned char *c);
int mg3_fifo_putc_unlocked (struct mg3_fifo *fifo, unsigned char c);
int mg3_fifo_seek_unlocked (struct mg3_fifo *fifo);
static size_t mg3_fifo_size_unlocked (struct mg3_fifo *fifo);
static void mg3_fifo_reset_unlocked (struct mg3_fifo *fifo);


static inline size_t mg3_fifo_capacity (const struct mg3_fifo *fifo)
{
	return fifo->size;
}


static inline size_t mg3_fifo_size_unlocked (struct mg3_fifo *fifo)
{
	return fifo->wp - fifo->rp;
}


static inline void mg3_fifo_reset_unlocked (struct mg3_fifo *fifo)
{
	fifo->wp = fifo->rp = 0;
}






#else
#warning "don't include kernel headers in userspace"
#endif // __KERNEL__


#endif // _MG3_RBUF_H
