/*
 *  mg3_sci.c
 *
 *	Copyright 2008-2009 
 *	Uchibori Laboratory, Numazu National College of Technology
 *	All rights reserved.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, version 2 of the License.
 */


//
// このドライバは
// Linux Device Drivers, Third Edition
// を参考に作成されている．この文書はオンラインで公開されている．
// http://oreilly.com/catalog/linuxdrive3/book/index.csp
//




#include <asm/io.h>
#include <asm/ioctls.h>
#include <asm/semaphore.h>
#include <asm/uaccess.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/ctype.h>
#include <linux/interrupt.h>
#include <linux/ioctl.h>
#include <linux/ioport.h>
#include <linux/poll.h>
#include <linux/spinlock.h>
#include <linux/termios.h>
#include <linux/types.h>
#include <linux/wait.h>
#include <linux/version.h>
#include <motor/sci_ctl.h>
#include "fifo.h"
#include "sci.h"

#include <linux/delay.h>



#define MAGIC 0x6D696C73


// 割り込み処理ルーチン（ISR)の実行中，割り込みを禁止する？
#if MG3_SCI_BUSY == 0
#define IRQ_FLAG 0
#else
#define IRQ_FLAG IRQF_DISABLED
#endif // MG3_SCI_BUSY




#define print_isr_info(level, msg, dev)                   \
	do {                                                  \
		printk(level "(%s) [ISR %u] %s \t"                \
			"I/O %x "                                     \
			"LSR:%02x "                                   \
			"IIR:%02x "                                   \
			"IER:%02x "                                   \
			"rb:%u wb:%u "                                \
			"CNT:%u\n",                                   \
			MG3_SCI_MODULE_NAME, (unsigned)dev->irq, msg, \
			(unsigned)dev->io_addr,                       \
			(unsigned)dev->lsr,                           \
			(unsigned)dev->iir,                           \
			(unsigned)dev->ier,                           \
			mg3_fifo_size_unlocked(&dev->rd_buf),         \
			mg3_fifo_size_unlocked(&dev->wr_buf),         \
			dev->n_intr);                                 \
	} while (0)



// デバッグ用にread/write時のメッセージを表示する．
#define PRINT_DEBUG_MESSAGE


#ifdef PRINT_DEBUG_MESSAGE

// デフォルトで，デバッグメッセージを表示する.
// これが設定されていない場合，表示を行うには
// ioctl(MG3_SCI_IOENDMSG)で表示を有効にする必要がある.
// #define PRINT_DEBUG_MESSAGE_DEFAULT

static inline void debug_print_isr (const char *msg, SciDevice *dev)
{
	if ( dev->opt & MG3_SCI_PRT_DEBUG_INFO ) {
		print_isr_info(KERN_INFO, msg, dev);
	}
}


static inline void debug_print_isr_rw (const char *msg, char c, SciDevice *dev)
{
	if ( dev->opt & MG3_SCI_PRT_DEBUG_INFO ) {
		printk(KERN_INFO "(%s) [ISR %u] %s 0x%02x(%c)\t"
			"I/O %x "
			"LSR:%02x "
			"IIR:%02x "
			"IER:%02x "
			"rb:%u wb:%u "
			"CNT:%u\n",
			MG3_SCI_MODULE_NAME, (unsigned)dev->irq,
			msg, (unsigned)(c & 0xff),
			isalnum(c) ? c : ' ',
			(unsigned)dev->io_addr,
			(unsigned)dev->lsr,
			(unsigned)dev->iir,
			(unsigned)dev->ier,
			mg3_fifo_size_unlocked(&dev->rd_buf),
			mg3_fifo_size_unlocked(&dev->wr_buf),
			dev->n_intr);
	}
}


static void debug_print (const char *msg, struct file *filp)
{
	SciDevice *dev = filp->private_data;

	if ( dev->opt & MG3_SCI_PRT_DEBUG_INFO ) {
		printk(KERN_INFO "(%s) [%s%u] %s \t"
			"I/O %x "
			"LSR:%02x "
			"IIR:%02x "
			"IER:%02x "
			"rb:%u wb:%u\n",
			MG3_SCI_MODULE_NAME,
			MG3_SCI_DEV_NAME,
			iminor(filp->f_dentry->d_inode),
			msg,
			(unsigned)dev->io_addr,
			(unsigned)dev->lsr,
			(unsigned)dev->iir,
			(unsigned)dev->ier,
			mg3_fifo_size_unlocked(&dev->rd_buf),
			mg3_fifo_size_unlocked(&dev->wr_buf));
	}
}


static void debug_print_rw (const char *msg, char c, struct file *filp)
{
	SciDevice *dev = filp->private_data;

	if ( dev->opt & MG3_SCI_PRT_DEBUG_INFO ) {
		printk(KERN_INFO "(%s) [%s%u] %s 0x%02x(%c)\t"
			"I/O %x "
			"LSR:%02x "
			"IIR:%02x "
			"IER:%02x "
			"rb:%u wb:%u\n",
			MG3_SCI_MODULE_NAME,
			MG3_SCI_DEV_NAME,
			iminor(filp->f_dentry->d_inode),
			msg, (unsigned)(c & 0xff),
			isalnum(c) ? c : ' ',
			(unsigned)dev->io_addr,
			(unsigned)dev->lsr,
			(unsigned)dev->iir,
			(unsigned)dev->ier,
			mg3_fifo_size_unlocked(&dev->rd_buf),
			mg3_fifo_size_unlocked(&dev->wr_buf));
	}
}


static void debug_print_sleep (const char *msg, unsigned n, struct file *filp)
{
	SciDevice *dev = filp->private_data;

	if ( dev->opt & MG3_SCI_PRT_DEBUG_INFO ) {
		printk(KERN_INFO "(%s) [%s%u] %s(%u)\t"
			"I/O %x "
			"LSR:%02x "
			"IIR:%02x "
			"IER:%02x "
			"rb:%u wb:%u\n",
			MG3_SCI_MODULE_NAME,
			MG3_SCI_DEV_NAME,
			iminor(filp->f_dentry->d_inode),
			msg, n,
			(unsigned)dev->io_addr,
			(unsigned)dev->lsr,
			(unsigned)dev->iir,
			(unsigned)dev->ier,
			mg3_fifo_size_unlocked(&dev->rd_buf),
			mg3_fifo_size_unlocked(&dev->wr_buf));
	}
}



#else
#define debug_print_isr(msg, dev, irq)
#define debug_print_isr_rw(msg, c, dev, irq)
#define debug_print(msg, filp)
#define debug_print_rw(msg, c, filp)
#define debug_print_sleep(msg, n, filp)
#endif // PRINT_DEBUG_MESSAGE


// write()に割り込み処理を使わない．
// #define POLL_WRITE



// デバイスのメジャー番号
static dev_t mg3_sci_major;




// Mg3SciDeviceの初期化．
// I/OポートアドレスとIRQのみを初期化している．
// 残りは必要に応じて，sci_module_init()やsci_open()で初期化する．
static SciDevice sci_dev[MG3_SCI_DEV_N_MINOR] = {
	{
		.io_addr = MG3_SCI0_IO_BASE,
		.irq = MG3_SCI0_IRQ,
	},
	{
		.io_addr = MG3_SCI1_IO_BASE,
		.irq = MG3_SCI1_IRQ,
	}
};





// File Operationsの宣言
static int sci_ioctl (struct inode*, struct file*, unsigned int, unsigned long);
static int sci_open (struct inode *inode, struct file *filp);
static unsigned int sci_poll (struct file *filp, poll_table *wait_table);
static int sci_release (struct inode *inode, struct file *filp);
static ssize_t sci_read (struct file *, char __user *, size_t , loff_t *);
static ssize_t sci_write (struct file*, const char __user*, size_t , loff_t*);
static int sci_fsync (struct file *, struct dentry *, int datasync);



// カーネルに登録するFile Operations構造体
static struct file_operations mg3_sci_fops = {
	.owner   = THIS_MODULE,
	.llseek  = no_llseek,
	.read    = sci_read,
	.write   = sci_write,
	.poll    = sci_poll,
	.ioctl   = sci_ioctl,
	.open    = sci_open,
	.release = sci_release,
	.fsync   = sci_fsync
};




//
// read/write用のセマフォを同時に取得/解放する．
// 順番を間違えるとデッドロックの原因となるため，関数として定義しておく．
//
static int lock_all_semaphore (struct file * filp, SciDevice *dev)
{
	if ( filp->f_flags & O_NONBLOCK ) {
		if ( down_trylock(&dev->wr_sem) ) {
			return -EAGAIN;
		}
		if ( down_trylock(&dev->rd_sem) ) {
			up(&dev->wr_sem);
			return -EAGAIN;
		}
	}
	else {
		if ( down_interruptible(&dev->wr_sem) ) {
			debug_print("INTERRUPTED BY SIGNAL", filp);
			return -ERESTARTSYS;
		}
		if ( down_interruptible(&dev->rd_sem) ) {
			debug_print("INTERRUPTED BY SIGNAL", filp);
			return -ERESTARTSYS;
		}
	}

	return 0;
}



static void unlock_all_semaphore (SciDevice *dev)
{
	up(&dev->rd_sem);
	up(&dev->wr_sem);
}



//
// read/write用のスピンロックを同時に取得/解放する．
// 順番を間違え(ry
//
#define lock_all_spinlock(dev, flags)         \
do {                                          \
	spin_lock_irqsave(&dev->wr_slock, flags); \
	spin_lock(&dev->rd_slock);                \
} while (0)


#define unlock_all_spinlock(dev, flags)            \
do {                                               \
	spin_unlock(&dev->rd_slock);                   \
	spin_unlock_irqrestore(&dev->wr_slock, flags); \
} while (0)




//
// リセット
//
static void reset (SciDevice *dev)
{
	int flags;

	lock_all_spinlock(dev, flags);

	mg3_fifo_reset_unlocked(&dev->rd_buf);
	mg3_fifo_reset_unlocked(&dev->wr_buf);

	dev->error = 0;
	dev->lsr = 0;
	dev->ier = (MG3_SCI_IER_RX | MG3_SCI_IER_ST);
	dev->iir = 0;
	outb(dev->lsr, MG3_SCI_LSR(dev->io_addr));
	outb(dev->ier, MG3_SCI_IER(dev->io_addr));


	unlock_all_spinlock(dev, flags);
}



//
//
// ISR
//
//


// エラーをコンソールに表示する.
static inline void print_isr_error (const char *msg, SciDevice *dev)
{
	print_isr_info(KERN_ERR, msg, dev);
}



// 送信割り込みを受け付けないように設定する．
static inline void disable_tx (SciDevice *dev)
{
	dev->ier &= ~(MG3_SCI_IER_TX);             // 送信割り込みを不許可にする．
	outb(dev->ier, MG3_SCI_IER(dev->io_addr)); // 割り込み許可要因を設定する．
}


// エラーが生じた際に，送受信を停止させる．
static inline void stop_rtx (SciDevice *dev)
{
	dev->ier &= ~(MG3_SCI_IER_RX);             // 受信割り込みを不許可にする．
	dev->ier &= ~(MG3_SCI_IER_TX);             // 送信割り込みを不許可にする．
	outb(dev->ier, MG3_SCI_IER(dev->io_addr)); // 割り込み許可要因を設定する．

	// スリープしているスレッドを起こす．
	wake_up_interruptible_sync(&dev->rd_queue);
	wake_up_interruptible_sync(&dev->wr_queue);
	debug_print_isr("WAKE UP (ERROR)", dev);
}



// ステータス(受信エラー)割り込み処理ルーチン
static inline void isr_st (SciDevice *dev)
{
	unsigned int err;
	static const unsigned int rx_errors
			= (MG3_SCI_LSR_OVR | MG3_SCI_LSR_PER | MG3_SCI_LSR_FER);	


	err = dev->lsr & rx_errors;		
	if ( err ) {
		// 受信エラーが起こった．
		dev->error |= err;
		print_isr_error("Recv ERROR", dev);
	}

	if ( dev->lsr & MG3_SCI_LSR_BRK ) {
		// ブレーク信号を受信した．
		dev->error |= MG3_SCI_LSR_BRK;
		print_isr_error("Recieve BREAK", dev);
	}

	if ( dev->error ) {
		// リセットされるまで，受信割り込みを受け付けない．
		// 送信処理もストップさせる．
		stop_rtx(dev);

		// LSRをクリア
		dev->lsr &= ~(rx_errors | MG3_SCI_LSR_BRK);
	}
}


// 受信割り込み処理ルーチン
static inline void isr_rx (SciDevice *dev, uint8_t c)
{
	if ( dev->error ) {
		return;
	}

	if ( dev->lsr & MG3_SCI_LSR_RXD ) {
		// エラー無しに受信できている．
		// FIFOバッファに保存する．
		if ( mg3_fifo_putc(&dev->rd_buf, c) < 0 ) {
			// FIFOバッファが満タンで保存することができない．
			dev->error |= MG3_SCI_LSR_OVR;
	
			// リセットされるまで，受信割り込みを受け付けない．
			// 送信処理もストップさせる．
			stop_rtx(dev);

			print_isr_error("Recv OVERFLOW", dev);
		}
		else {
			debug_print_isr_rw("Recieved", c, dev);

			// 受信完了フラグ(MG3_SCI_LSR_RXD)をクリアする．
			dev->lsr &= ~MG3_SCI_LSR_RXD;

			// データ待ちで眠っているスレッドを起こす．
			if ( mg3_fifo_size(&dev->rd_buf) >= dev->n_rd_wakeup ) {
				wake_up_interruptible_sync(&dev->rd_queue);
				debug_print_isr("WAKE UP (read)", dev);
			}
		}
	}
}



static void isr_send (SciDevice *dev)
{
	char c;

	if ( mg3_fifo_getc_unlocked(&dev->wr_buf, &c) == 0 ) {
		debug_print_isr_rw("Sending", c, dev);

		// 送るべきデータがあるので，送信レジスタに転送する．
		dev->lsr &= ~(MG3_SCI_LSR_TXR);
		outb(c, MG3_SCI_THR(dev->io_addr));        // 送信
		outb(dev->lsr, MG3_SCI_LSR(dev->io_addr)); // 送信ステータスをクリア
		debug_print_isr_rw("Sent", c, dev);
	}
	else {
	}
}


// 送信割り込み処理ルーチン
static inline void isr_tx (SciDevice *dev)
{
	unsigned long flags;


	if ( dev->error ) {
		return;
	}

	spin_lock_irqsave(&dev->wr_slock, flags);

	if ( dev->lsr & MG3_SCI_LSR_TXR ) {
		// 送信レジスタが空なので，
		// ここで送信データを送信レジスタに転送する．
		isr_send(dev);

		// 空き待ちで眠っているスレッドを起こす．
		if ( mg3_fifo_size_unlocked(&dev->wr_buf) <= dev->n_wr_wakeup ) { 
			wake_up_interruptible_sync(&dev->wr_queue);
			debug_print_isr("WAKE UP (write)", dev);
		}
	}


	if ( dev->lsr & MG3_SCI_LSR_TXD ) {
		if ( mg3_fifo_size_unlocked(&dev->wr_buf) != 0 ) {
			panic("mg3_sci: FIFO is not empty but dev->lsr & MG3_SCI_LSR_TXD != 0.\n");
		}

		// この時点で送信中ではない場合，
		// 送るデータはもうないはずなので，
		// 割り込みを停止させる．
		if ( (dev->lsr & MG3_SCI_LSR_TXR) != 0 ) {
			debug_print_isr("Send Completely", dev);

			// 送信用FIFOバッファが空(i.e. もう送るデータがない)なので
			// 送信割り込みを不許可にして，送信処理を終わらせる．
			disable_tx(dev);

			// ついでにFIFOバッファをリセットする．
			mg3_fifo_reset_unlocked(&dev->wr_buf);

			// 送信完了待ちで眠っているスレッドを起こす．		
			wake_up_interruptible_sync(&dev->sy_queue);
			debug_print_isr("WAKE UP (sync)", dev);
		}
		dev->lsr &= ~(MG3_SCI_LSR_TXD);
		outb(dev->lsr, MG3_SCI_LSR(dev->io_addr)); // 送信ステータスをクリア
	}

	spin_unlock_irqrestore(&dev->wr_slock, flags);
}



// Ver. 2.6.19からISRのプロトタイプが変更されている．
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 19)
static irqreturn_t isr (int irq, void *arg, struct pt_regs *regs)
#else
static irqreturn_t isr (int irq, void *arg)
#endif
{
	uint8_t c;   // 読み込み用
	uint8_t lsr; // ラインステータスのクリア用
	SciDevice *dev = arg;


	// ラインステータスより前に，読み込みデータを保存しないと，
	// オーバーランエラーを読み落とす可能性があるため，
	// まず，データを読み込む．
	c = inb(MG3_SCI_RBR(dev->io_addr));


	// 割り込み要因を保存する．
	dev->iir = inb(MG3_SCI_IIR(dev->io_addr)) & MG3_SCI_IIR_MASK;


	// ラインステータスを保存する．
	dev->lsr = inb(MG3_SCI_LSR(dev->io_addr));

	// ラインステータスをクリアする．
	// ブレークと送信に関するフラグはそのままにしておく．
	// ラインステータスを保存してからクリアするまでの間に
	// 状態の変化(データを受信したり，受信エラーが起こったり)があると，
	// その情報を読み落とすことになる．どうしよう？？？
	lsr = (dev->lsr & (MG3_SCI_LSR_BSK | MG3_SCI_LSR_BRK | MG3_SCI_LSR_TXR | MG3_SCI_LSR_TXD));
	outb(lsr, MG3_SCI_LSR(dev->io_addr));

	++dev->n_intr;
	debug_print_isr("Detect INTR.", dev);

	switch ( dev->iir ) {
	case (MG3_SCI_IIR_ST) : // 受信エラー処理
		isr_st(dev);
		// FALLTHROUGH

	case (MG3_SCI_IIR_RX) : // 受信処理
		isr_rx(dev, c);
		// FALLTHROUGH

	case (MG3_SCI_IIR_TX): // 送信処理
#ifndef POLL_WRITE
		isr_tx(dev);
#endif  // POLL_WRITE
		// FALLTHROUGH

	case (MG3_SCI_IIR_NONE): // 割り込みなし
		dev->iir = 0;
		return IRQ_HANDLED;
		break;

	default:
		return IRQ_NONE;
		break;
	}
}





//
//
// File Operations
//
//


//
// fsync()
//

#ifdef POLL_WRITE

static int sci_fsync (struct file *filp, struct dentry *dent, int datasync)
{
	SciDevice *dev = filp->private_data;


	if ( dev->error != 0 ) {
		printk(KERN_INFO "(%s) [%s%u] READ ERROR:%02x (found in fsync process)\n", 
				MG3_SCI_MODULE_NAME, MG3_SCI_DEV_NAME, 
				iminor(filp->f_dentry->d_inode), dev->error);
		return -EIO;
	}

	while ( dev->lsr & MG3_SCI_LSR_TXR ) {
		msleep_interruptible(1);
	}

	return 0;
}


#else

static int sci_fsync (struct file *filp, struct dentry *dent, int datasync)
{
	SciDevice *dev = filp->private_data;
	int retval = 0;


	// open()されていないのにfsync()が呼ばれたので，処理を中断する．
	if ( dev->magic != MAGIC ) {
		return -EBADF;
	}


	// 読み込み時にデバイスがエラーを起こしている場合は，直ちに戻る．
	if ( dev->error != 0 ) {
		printk(KERN_INFO "(%s) [%s%u] READ ERROR:%02x (found in fsync process)\n", 
				MG3_SCI_MODULE_NAME, MG3_SCI_DEV_NAME, 
				iminor(filp->f_dentry->d_inode), dev->error);
		return -EIO;
	}


	// セマフォで他の書き込みスレッドをブロックする．
	if ( down_interruptible(&dev->wr_sem) ) {
		debug_print("INTERRUPTED BY SIGNAL (FSYNC)", filp);
		return -ERESTARTSYS;
	}


	// 書き込み用FIFOバッファが空になるまで待つ．
	while ( mg3_fifo_size(&dev->wr_buf) > 0 ) {
		DEFINE_WAIT(node);
		debug_print_sleep("SLEEP TO FSYNC", 0, filp);
		prepare_to_wait(&dev->sy_queue, &node, TASK_INTERRUPTIBLE);
		if ( mg3_fifo_size(&dev->wr_buf) > 0 ) {
			schedule();
		}
		finish_wait(&dev->sy_queue, &node);
		if ( signal_pending(current) ) {
			// シグナルに割り込まれた．
			// すでに読み込まれたデータがあれば，その数を返す．
			// なければ，上位層にお任せする．
			debug_print("INTERRUPTED BY SIGNAL (FSYNC)", filp);
			retval = -ERESTARTSYS;
			break;
		}
	}

	up(&dev->wr_sem);

	return retval;
}

#endif // POLL_WRITE



//
// ioctl()
//
static inline void break_on (SciDevice *dev)  // down
{
	dev->lsr |= MG3_SCI_LSR_BSK;
	outb(0xff, MG3_SCI_LSR(dev->io_addr));
}



static inline void break_off (SciDevice *dev) // up
{
	dev->lsr &= ~(MG3_SCI_LSR_BSK);
	outb(0x7f, MG3_SCI_LSR(dev->io_addr));
}



static int sci_ioctl (struct inode *inode, struct file *filp,
						unsigned int cmd, unsigned long arg)
{
	SciDevice *dev = filp->private_data;
	int retval = 0;


	retval = lock_all_semaphore(filp, dev);
	if ( retval ) {
		return retval;
	}


	if ( _IOC_TYPE(cmd) == SCI_IOCTL_TYPE ) {
		//
		// 独自定義
		//
		switch ( cmd ) {

		// LSR中のエラーコードを取得する．
		case MG3_SCI_IOCGLSR:
			retval = dev->error & 0x00ff;
			break;

		// LSR中のエラーコードをリセットする．
		case MG3_SCI_IOCCLSR:
			dev->error = 0;
			break;

		// デバイスをリセットする.
		case MG3_SCI_IOCRESET:
			reset(dev);
			break;

		// デバッグ用のメッセージを有効にする．
		case MG3_SCI_IOENDMSG:
			dev->opt |= MG3_SCI_PRT_DEBUG_INFO;
			break;
			
		// デバッグ用のメッセージを無効にする．
		case MG3_SCI_IODSDMSG:
			dev->opt &= ~MG3_SCI_PRT_DEBUG_INFO;
			break;

		default:
			retval = -ENOTTY;
			break;
		}
	}
	else {
		//
		// tty
		//
		switch ( cmd ) {
		// ブレーク信号送出
		case TIOCSBRK:
			break_on(dev);
			break;

		// ブレーク信号停止
		case TIOCCBRK:
			break_off(dev);
			break;

		// 入力バッファにあるバイト数
		case TIOCINQ: //FIONREAD:
			retval = put_user(mg3_fifo_size(&dev->rd_buf), (int __user *)arg);
			break;			

		// 出力バッファにあるバイト数
		case TIOCOUTQ:
			retval = put_user(mg3_fifo_size(&dev->wr_buf), (int __user *)arg);
			break;

		// バッファの破棄
		case TCFLSH:
			switch ( arg ) {
			case TCIFLUSH:
				mg3_fifo_reset(&dev->rd_buf);
				debug_print("Flush FIFO (TCIFLUSH)", filp);
				break;
			case TCOFLUSH:
				mg3_fifo_reset(&dev->wr_buf);
				debug_print("Flush FIFO (TCOFLUSH)", filp);
				break;
			case TCIOFLUSH:
				mg3_fifo_reset(&dev->rd_buf);
				mg3_fifo_reset(&dev->wr_buf);
				debug_print("Flush FIFO (TCIOFLUSH)", filp);
				break;
			default:
				retval = -EINVAL;
				debug_print("Flush FIFO (UNKNOWN)", filp);
				break;
			}
			break;

		default:
			retval = -ENOTTY;
			break;
		}
	}

	unlock_all_semaphore(dev);

	return retval;
//	return -ENOIOCTLCMD;
}



//
// open()
//
static int sci_open (struct inode *inode, struct file *filp)
{
	SciDevice *dev;
	struct resource *io_resource;
	int err = 0;


	dev = container_of(inode->i_cdev, SciDevice, cdev);


	// セマフォを取得
	err = lock_all_semaphore(filp, dev);
	if ( err ) {
		return err;
	}



	// 重複したオープンの処理
	if ( dev->magic != 0 ) {
		// 既に誰かがopenしているかその最中なので失敗させる
		unlock_all_semaphore(dev);
		return -EBUSY;
	}

	// いちいちcontainer_of()をしなくてよいように
	// private_dataにdevを記録しておく．
	filp->private_data = dev;


	// Read/Write Bufferを初期化する．
	BUG_ON(dev->rd_buf.buffer != NULL);
	err = mg3_fifo_init(&dev->rd_buf, MG3_SCI_FIFO_SIZE, GFP_KERNEL, &dev->rd_slock);
	if ( err !=  0 ) {
		goto allocate_rd_buf_fail;
	}
	BUG_ON(dev->wr_buf.buffer != NULL);
	err = mg3_fifo_init(&dev->wr_buf, MG3_SCI_FIFO_SIZE, GFP_KERNEL, &dev->wr_slock);
	if ( err !=  0 ) {
		goto allocate_wr_buf_fail;
	}


	// データ待ち用のQueueを初期化する．
	init_waitqueue_head(&dev->rd_queue);
	init_waitqueue_head(&dev->wr_queue);
	init_waitqueue_head(&dev->sy_queue);


	// I/O Portアドレスを割り当てる．
	io_resource = request_region(dev->io_addr, MG3_SCI_IO_SIZE, MG3_SCI_MODULE_NAME);
	if ( io_resource == NULL ) {
		printk(KERN_INFO "(%s) [%s%u] fail to allocate I/O port %x - %x.\n",
					MG3_SCI_MODULE_NAME,
					MG3_SCI_DEV_NAME, iminor(inode),
					(unsigned)dev->io_addr, (unsigned)dev->io_addr + MG3_SCI_IO_SIZE - 1);
		err = -EBUSY;
		goto allocate_io_fail;
	}
	printk(KERN_INFO "(%s) [%s%u] I/O port : %x - %x\n",
				MG3_SCI_MODULE_NAME,
				MG3_SCI_DEV_NAME, iminor(inode),
				(unsigned)dev->io_addr, (unsigned)dev->io_addr + MG3_SCI_IO_SIZE - 1);


	// FIFOバッファや，ライン・割り込み許可状態等をリセットする．
	reset(dev);


	// lseek()できないデバイスです．
	nonseekable_open(inode, filp);


	// IRQを割り当てる．
	err = request_irq(dev->irq, isr, IRQ_FLAG, MG3_SCI_MODULE_NAME, dev);
	if ( err != 0 ) {
		printk(KERN_INFO "(%s) [%s%u] fail to allocat IRQ %u.\n",
					MG3_SCI_MODULE_NAME, MG3_SCI_DEV_NAME,
					iminor(inode), (unsigned)dev->irq);
		goto allocate_irq_fail;
	}
	printk(KERN_INFO "(%s) [%s%u] IRQ : %u\n",
				MG3_SCI_MODULE_NAME, MG3_SCI_DEV_NAME,
				iminor(inode), (unsigned)dev->irq);


	dev->opt = 0;
#ifdef PRINT_DEBUG_MESSAGE
#ifdef PRINT_DEBUG_MESSAGE_DEFAULT
	// デバッグ用メッセージの制御
	dev->opt |= MG3_SCI_PRT_DEBUG_INFO;
#endif // PRINT_DEBUG_MESSAGE_DEFAULT
#endif // PRINT_DEBUG_MESSAGE


	// 割り込み回数をリセット
	dev->n_intr = 0;


	// 正常にオープンしたことを記録する．
	dev->magic = MAGIC;
	// バリアが必要?


	// ブレーク状態を解除する．
	break_off(dev);


	printk(KERN_INFO "(%s) [%s%d] Opend.\n",
				MG3_SCI_MODULE_NAME,
				MG3_SCI_DEV_NAME, iminor(inode));

	unlock_all_semaphore(dev);

	return 0; // 成功



	//
	// openに失敗したときの後始末．
	//
allocate_irq_fail:
	// I/Oポートを解放する．
	release_region(dev->io_addr, MG3_SCI_IO_SIZE);

allocate_io_fail:
	// 書き込み用FIFOバッファを解放する．
	mg3_fifo_destroy(&dev->wr_buf);

allocate_wr_buf_fail:
	// 読み込み用FIFOバッファを解放する．
	mg3_fifo_destroy(&dev->rd_buf);

allocate_rd_buf_fail:
	dev->magic = 0;
	unlock_all_semaphore(dev);
	return err;

}



//
// select / poll
//
static unsigned int sci_poll (struct file *filp, poll_table *wait_table)
{
	SciDevice *dev = filp->private_data;
	unsigned int mask = 0;
	int flags;


	// データ(の空き)があれば，すぐに起こす.
	dev->n_rd_wakeup = 0;
	dev->n_wr_wakeup = MG3_SCI_FIFO_SIZE - 1;

	lock_all_spinlock(dev, flags);

	poll_wait(filp, &dev->rd_queue, wait_table);
	poll_wait(filp, &dev->wr_queue, wait_table);


	// ブロックされずにread()できる？
	if ( mg3_fifo_size_unlocked(&dev->rd_buf) > 0 ) {
		// FIFOバッファにデータがあるので，ブロックなしで読み出し可能
		mask |= POLLIN | POLLRDNORM;
		debug_print("Poll:Read OK", filp);
	}

	// ブロックされずにwrite()できる？
#ifdef POLL_WRITE
	// 割り込みなしのwrite()がブロックされるのはセマフォのみ．
	// ロックが成功すれば，書き込み可能だと判断する．
	if ( down_trylock(&dev->wr_sem) == 0 ) {
		mask |= POLLOUT | POLLWRNORM;
		up(&dev->wr_sem);
		debug_print("Poll:Write OK", filp);
	}
#else	
	if ( mg3_fifo_size_unlocked(&dev->wr_buf) < mg3_fifo_capacity(&dev->wr_buf) ) {
		// FIFOバッファに空きがあるので，ブロックなしで書き込み可能
		mask |= POLLOUT | POLLWRNORM;
		debug_print("Poll:Write OK", filp);
	}
#endif

	// エラーが起こってる？
	if ( dev->error != 0 ) {
		// エラーが起きている．
		mask |= POLLERR;
		debug_print("Poll:Some Err", filp);
	}


	unlock_all_spinlock(dev, flags);

	return mask;
}



//
// release()
//
static int sci_release (struct inode *inode, struct file *filp)
{
	SciDevice *dev = filp->private_data;


	// open()されていないのにclose()が呼ばれたので，処理を中断する．
	if ( dev->magic != MAGIC ) {
		return -EBADF;
	}

	// 送受信割り込みを不許可にする．
	dev->ier = 0;
	outb(dev->ier, MG3_SCI_IER(dev->io_addr));


	// IRQを解放する．
	free_irq(dev->irq, dev);


	// Portアドレスを解放する．
	release_region(dev->io_addr, MG3_SCI_IO_SIZE);


	// Read/Write Bufferを解放する．
	mg3_fifo_destroy(&dev->rd_buf);
	mg3_fifo_destroy(&dev->wr_buf);

	filp->private_data = NULL;

	// 既にオープンされていないことを記録する．
	dev->magic = 0;

	printk(KERN_INFO "(%s) [%s%u] Closed.\n",
			MG3_SCI_MODULE_NAME, MG3_SCI_DEV_NAME, iminor(inode));

	return 0;
}



//
// read()
//

// FIFOバッファから1文字読み込む．
// 成功したら，0が返る．
static inline ssize_t read_from_fifo (struct file *filp, uint8_t *c, size_t leftover)
{
	SciDevice *dev = filp->private_data;

	// FIFOバッファからの読み込みに成功した後，
	// ユーザ指定のバッファへの書き込みに失敗すると
	// 読み込んだデータをFIFOバッファに戻すことが
	// (FIFOバッファへの書き込みの可能性があるので)
	// 確実には出来ないため，このデータをロストしてしまう可能性がある．
	// これを防ぐため，ここではFIFOバッファをシークさせずに
	// データの内容を見るだけのpeek()を用いて，
	// シークはユーザ指定のバッファへの書き込みに成功した後に行う．
	while ( mg3_fifo_peek(&dev->rd_buf, c) < 0 ) {
		// FIFOバッファが空で，読み込みに失敗した．

		if ( dev->error != 0 ) {
			// ハードウェアから通信(読み込み)エラーが報告されている．
			printk(KERN_INFO "(%s) [%s%u] READ ERROR:%02x\n", 
					MG3_SCI_MODULE_NAME, MG3_SCI_DEV_NAME, 
					iminor(filp->f_dentry->d_inode), dev->error);
			return -EIO;
		}
		else if ( filp->f_flags & O_NONBLOCK ) {
			// 非ブロック処理が指定されているので，直ちに戻る．
			return -EAGAIN;
		}
		else {
			//　ブロック処理が指定されているので，スリープする．
			DEFINE_WAIT(node);

			// 起こしてもらうデータ数を指定しておく．
			dev->n_rd_wakeup =
				(leftover < MG3_SCI_RD_WAKEUP_MAX ? leftover : MG3_SCI_RD_WAKEUP_MAX);
			debug_print_sleep("SLEEP TO READ", dev->n_rd_wakeup, filp);

			// スリープ
			prepare_to_wait(&dev->rd_queue, &node, TASK_INTERRUPTIBLE);
			if ( mg3_fifo_size(&dev->rd_buf) == 0 ) {
				// GGG...
				schedule();
			}
			finish_wait(&dev->rd_queue, &node);
			if ( signal_pending(current) ) {
				// シグナルに割り込まれた．
				debug_print("INTERRUPTED BY SIGNAL (READ)", filp);
				return -ERESTARTSYS;
			}
		}
	} // end of FIFOバッファからの読み込み

	return 0;
}



static ssize_t sci_read (struct file *filp, char __user *p, size_t size, loff_t *offset)
{
	SciDevice *dev = filp->private_data;
	size_t n_read = 0; // 読み込まれたデータの数
	int err;
	unsigned char c;

	// open()されていないのにread()が呼ばれたので，処理を中断する．
	if ( dev->magic != MAGIC ) {
		return -EBADF;
	}

	// ユーザの示したバッファのアドレス範囲が有効か調べる．
	if ( !access_ok(VERIFY_WRITE, p, size) ) {
		// アクセスできません．
		return -EFAULT;
	}

	// セマフォで他の読み込みスレッドをブロックする．
	if ( filp->f_flags & O_NONBLOCK ) {
		if ( down_trylock(&dev->rd_sem) ) {
			return -EAGAIN;
		}
	}
	else if ( down_interruptible(&dev->rd_sem) ) {
		debug_print("INTERRUPTED BY SIGNAL (READ)", filp);
		return -ERESTARTSYS;
	}

	// 指定のデータ数だけ，FIFOから読み込み，ユーザ指定のバッファpに書き込む.
	while ( n_read < size ) {
		// FIFOバッファから1文字読み込む．
		err = read_from_fifo(filp, &c, size - n_read);
		if ( err < 0 ) {
			goto sci_read_error;
		}

		// ユーザ指定のバッファに書き込む．
		err = __put_user(c, p);
		if ( err < 0 ) {
			// ユーザ領域への書き込みに失敗した．
			// デバイスから読み込まれたデータがあれば，その数を返す．
			// なければ，そのままエラーコードを返す．
			goto sci_read_error;
		}

		// 読み込みに成功した．
		++p;
		++n_read;
		mg3_fifo_seek(&dev->rd_buf); // FIFOバッファの読み込み位置を次の文字位置にする．
		debug_print_rw("Read", c, filp);
	} // end of "while ( n_read < size )"


	// もう起こしてもらわなくてもよいので，閾値データ数は'MG3_SCI_RD_WAKEUP_MAX'にする．
	dev->n_rd_wakeup = MG3_SCI_RD_WAKEUP_MAX;


	up(&dev->rd_sem);

	return size;

sci_read_error:
	// 読み込み時にエラー(ハードウェア，シグナル割り込み，非ブロック読み込みでデータなし)が発生した．
	// すでに読み込まれたデータがあればその数を返し，
	// なければエラーコードを返す．
	up(&dev->rd_sem);
	return ( n_read > 0 ) ? n_read : err;
}



#ifdef POLL_WRITE

static inline ssize_t send_tx (struct file *filp, uint8_t c)
{
	SciDevice *dev = filp->private_data;


	if ( dev->error != 0 ) {
		printk(KERN_INFO "(%s) [%s%u] READ ERROR:%02x (found in write process)\n", 
				MG3_SCI_MODULE_NAME, MG3_SCI_DEV_NAME, 
				iminor(filp->f_dentry->d_inode), dev->error);
		return -EIO;
	}

	if ( dev->lsr & MG3_SCI_LSR_TXR ) {
		outb(c, MG3_SCI_THR(dev->io_addr));
		debug_print_rw("Send TX", c, filp);
		return 0;
	}

	return -EAGAIN;
}



static ssize_t sci_write (struct file *filp, const char __user *p, size_t size, loff_t *offset)
{
	ssize_t err;
	size_t i;
	unsigned char c;
	SciDevice *dev = filp->private_data;


	// セマフォで他のスレッドとの競合をさける．
	if ( filp->f_flags & O_NONBLOCK ) {
		if ( down_trylock(&dev->rd_sem) ) {
			return -EAGAIN;
		}
	}
	else if ( down_interruptible(&dev->wr_sem) ) {
		debug_print("INTERRUPTED BY SIGNAL (WRITE)", filp);
		return -ERESTARTSYS;
	}

	dev->lsr = inb(MG3_SCI_LSR(dev->io_addr));

	for ( i = 0; i < size; i++ ) {
		// ユーザ指定のバッファから読み込む.
		err = get_user(c, p);
		if ( err < 0 ) {
			up(&dev->wr_sem);
			return err;
		}

		// デバイスへ1文字書き込む．
		err = send_tx(filp, c);
		if ( err == -EAGAIN ) {
			// 書き込むスペースがない
			msleep_interruptible(1);
			--i;
			continue;
		}
		if ( err < 0 ) {
			up(&dev->wr_sem);
			return err;
		}
		++p;
		msleep_interruptible(1);
	}

	up(&dev->wr_sem);

	return i;
}


#else

// 割り込みを用いてwrite()


// 送信バッファにデータを送る．
static inline void send_tx (SciDevice *dev, uint8_t c)
{
	dev->ier |= MG3_SCI_IER_TX;
	dev->lsr &= ~(MG3_SCI_LSR_TXR);
	outb(dev->ier, MG3_SCI_IER(dev->io_addr)); // 送信割り込み許可
	outb(c, MG3_SCI_THR(dev->io_addr));        // 送信
	outb(dev->lsr, MG3_SCI_LSR(dev->io_addr)); // 送信ステータスをクリア
}



// 送信処理が始められていなければ，
// FIFOバッファから1文字取り出し，送信を開始する.
// 戻り値は送信した文字数(送信しなければ0)．
static ssize_t start_tx (struct file *filp)
{
	SciDevice *dev = filp->private_data;
	unsigned long flags; // 割り込み許可情報
	ssize_t retval = 0;
	char c;


	if ( dev->error ) {
		// 受信エラーが発生しているので，送信も行わない．
		return 0;
	}

	spin_lock_irqsave(&dev->wr_slock, flags);

	// 送信処理が行われている間は送信割り込みが許可されているので，
	// IERレジスタのMG3_SCI_IER_TXビットを見て送信中かどうかを判断する.
	if ( (dev->ier & MG3_SCI_IER_TX) == 0 ) {
		// 送信処理は行われていないので，送信を開始する．
		// FIFOバッファから一文字取り出して（取り出せたら）
		// 送信割り込みを許可し，送信バッファに書き込む．
		if ( mg3_fifo_getc_unlocked(&dev->wr_buf, &c) == 0 ) {
			send_tx(dev, c);
			retval = 1;
			debug_print_rw("Start TX", c, filp);
		}
		else {
			// FIFOバッファが空なので，(カウンターがオーバーフローしないように)リセットしておく．
			mg3_fifo_reset_unlocked(&dev->wr_buf);
		}
	}

	spin_unlock_irqrestore(&dev->wr_slock, flags);

	return retval;
}


// 送信処理中ではなく，要求送信文字数が1のときは，
// FIFOバッファを通さずに，直接送信バッファに送る．
// 戻り値は送信した文字数(送信しなければ0)かエラーコード．
static inline ssize_t write_1 (struct file *filp, const char __user *p)
{
	SciDevice *dev = filp->private_data;
	ssize_t err;
	char c;


	// 送信処理が行われていれば，何もしない.
	if ( (dev->ier & MG3_SCI_IER_TX) != 0 ) {
		return 0;
	}

	// ユーザ指定のバッファpから読み込む.
	err = __get_user(c, p);
	if ( err < 0 ) {
		// ユーザ領域からの読み込みに失敗した．
		// エラーコードをそのまま返す．
		return err;
	}
	else {
		// 送信バッファに書き込む．
		send_tx(dev, c);
		debug_print_rw("Write_1", c, filp);

		return 1;
	}
}


// 書き込み用FIFOバッファが空くのを待っているスレッドを起こすデータ数の閾値をセットする．
static void inline set_wr_wakeup (SciDevice *dev, size_t leftover)
{
	size_t n;

	if ( leftover > MG3_SCI_FIFO_SIZE ) {
		leftover = MG3_SCI_FIFO_SIZE;
	}
	n = MG3_SCI_FIFO_SIZE - leftover;
	if ( MG3_SCI_WR_WAKEUP_MIN > n ) {
		n = MG3_SCI_WR_WAKEUP_MIN;
	}
	dev->n_wr_wakeup = n;
}



// FIFOバッファに1文字書き込む．
// 成功したら，0を返す．
static inline ssize_t write_to_fifo (struct file *filp, uint8_t c, size_t leftover)
{
	SciDevice *dev = filp->private_data;

	while ( mg3_fifo_putc(&dev->wr_buf, c) < 0 ) {
		// FIFOバッファがいっぱいで，書き込みに失敗した．

		// 受信エラーが発生しているので，送信も行わない．
		if ( dev->error != 0 ) {
			printk(KERN_INFO "(%s) [%s%u] READ ERROR:%02x (found in write process)\n", 
					MG3_SCI_MODULE_NAME, MG3_SCI_DEV_NAME, 
					iminor(filp->f_dentry->d_inode), dev->error);
			return -EIO;
		}

		// 送信を開始
		if ( start_tx(filp) > 0 ) {
			// 送信が，たった今，開始された．
			// FIFOバッファが1文字分きっと空いている．
			continue;
		}

		// FIFOバッファが空くのを待って，スリープする？
		if ( filp->f_flags & O_NONBLOCK ) {
			// 非ブロック処理が指定されているので，直ちに戻る．
			return -EAGAIN;
		}
		else {
			//　ブロック処理が指定されているので，スリープする．
			DEFINE_WAIT(node);
			set_wr_wakeup(dev, leftover);
			debug_print_sleep("SLEEP TO WRITE", dev->n_wr_wakeup, filp);
			prepare_to_wait(&dev->wr_queue, &node, TASK_INTERRUPTIBLE);
			if ( mg3_fifo_size(&dev->wr_buf) == MG3_SCI_FIFO_SIZE ) {
				schedule();
			}
			finish_wait(&dev->wr_queue, &node);
			if ( signal_pending(current) ) {
				// シグナルに割り込まれた．
				// すでに読み込まれたデータがあれば，その数を返す．
				// なければ，上位層にお任せする．
				debug_print("INTERRUPTED BY SIGNAL (WRITE)", filp);
				return -ERESTARTSYS;
			}
		}
	} // end of FIFOバッファへの書き込み

	return 0;
}


//
// write()
//
static ssize_t sci_write (struct file *filp, const char __user *p, size_t size, loff_t *offset)
{
	SciDevice *dev = filp->private_data;
	size_t n_written = 0; // 書き込まれたデータの数
	int err;
	char c;

	// open()されていないのにwrite()が呼ばれたので，処理を中断する．
	if ( dev->magic != MAGIC ) {
		return -EBADF;
	}

	// 受信エラーが発生しているので，送信も行わない．
	// select/poll後の呼び出しの可能性があるので直ちに戻る．
	if ( dev->error != 0 ) {
		printk(KERN_INFO "(%s) [%s%u] READ ERROR:%02x (found in write process)\n", 
				MG3_SCI_MODULE_NAME, MG3_SCI_DEV_NAME, 
				iminor(filp->f_dentry->d_inode), dev->error);
		return -EIO;
	}

	// ユーザの示したバッファのアドレス範囲が有効か調べる．
	if ( !access_ok(VERIFY_READ, p, size) ) {
		// アクセスできません．
		return -EFAULT;
	}


	// セマフォで他の書き込みスレッドをブロックする．
	if ( filp->f_flags & O_NONBLOCK ) {
		if ( down_trylock(&dev->rd_sem) ) {
			return -EAGAIN;
		}
	}
	else if ( down_interruptible(&dev->wr_sem) ) {
		debug_print("INTERRUPTED BY SIGNAL (WRITE)", filp);
		return -ERESTARTSYS;
	}


	// 送信するデータが1つのみの場合，直接送信バッファに送ってみる．
	if ( size == 1 ) {
		ssize_t result = write_1(filp, p);
		if ( result != 0 ) {
			up(&dev->wr_sem);
			return result;
		}
	}

	// 指定のデータ数だけ，ユーザ指定のバッファpから読み込み，FIFOバッファに書き込む.
	// デバイスが送信処理をしていない場合，処理を始めさせる．
	while ( n_written < size ) {
		// ユーザ指定のバッファpから読み込む.
		err = __get_user(c, p);
		if ( err < 0 ) {
			// ユーザ領域からの読み込みに失敗した．
			goto sci_write_error;
		}

		// FIFOバッファに書き込む．
		err = write_to_fifo(filp, c, size - n_written);
		if ( err < 0 ) {
			// FIFOへの書き込みに失敗した．
			goto sci_write_error;
		}

		// FIFOバッファへの書き込みに成功した．
		// 次のデータに備えて調整．
		++p;
		++n_written;

		debug_print_rw("Write", c, filp);
	} // end of "while ( n_remain )"


	// 送信を開始する.
	start_tx(filp);

	// もう起こしてもらわなくてもよいので，閾値データ数は0にする．
	dev->n_wr_wakeup = 0;


	up(&dev->wr_sem);


	return size;


sci_write_error:
	// 何かエラーが発生した．
	// デバイスに書かれたデータがあれば，その数を返す．
	// なければ，エラーコードをそのまま返す．
	up(&dev->wr_sem);
	return (n_written > 0) ? n_written : err;
}


#endif // POLL_WRITE



//
//
// Initialize and Release
//
//


// デバイスのメジャー番号を割り当てる．
// Linux Device Drivers, Third Edition
// p.45 "Allocating and Freeing Device Numbers"を参照
static inline int allocate_device_number (unsigned int major, unsigned int nr_dev, char *name)
{
	int result;
	dev_t dev;

	if ( major == 0 ) {
		// デバイスのメジャー番号を自動で割り当てる，
		result = alloc_chrdev_region(&dev, 0, nr_dev, name);
		major = MAJOR(dev);
	}
	else {
		// デバイスのメジャー番号を手動で割り当てる，
		dev = MKDEV(major, 0);
		result = register_chrdev_region(dev, nr_dev, name);
	}

	if ( result < 0 ) {
		printk(KERN_WARNING "%s : cannot get device major %d\n", name, major);
		return result;
	}
	else {
		mg3_sci_major = major;
		return major;
	}
}



// kernelが使用するcdev構造体を初期化する．
// Linux Device Drivers, Third Edition
// p.55 "Char Device Registration"を参照．
static void setup_cdev (struct cdev *cdev)
{
	cdev_init(cdev, &mg3_sci_fops);
	cdev->owner = THIS_MODULE;
	cdev->ops = &mg3_sci_fops;
}



// モジュールを初期化する．
static int sci_module_init (void)
{
	int err = 0;
	int major, minor;
	struct cdev *cdev;
	SciDevice *dev;
	dev_t devno;



	// デバイスのメジャー番号を割り当てる．
	major = allocate_device_number(
			MG3_SCI_MAJOR, MG3_SCI_DEV_N_MAJOR, MG3_SCI_DEV_NAME);
	if ( major < 0 ) {
		printk(KERN_INFO "%s : fail to allocate MG3 SCI device number.\n",
				MG3_SCI_DEV_NAME);
		return major; // major is error code.
	}


	for ( minor = 0; minor < MG3_SCI_DEV_N_MINOR; ++minor ) {
		dev = &sci_dev[minor];
		cdev = &dev->cdev;

		// sci_devのspinlockとsemaphoreを初期化する．
		// sci_devの残りのデータは，open()時に初期化する．
		spin_lock_init(&dev->rd_slock);
		spin_lock_init(&dev->wr_slock);
		sema_init(&dev->rd_sem, 1);
		sema_init(&dev->wr_sem, 1);

		// kernelが使用するcdev構造体を初期化する．
		setup_cdev(cdev);

		// カーネルにデバイスを登録する．
		// 登録後，直ちにデバイスは使用できる状態になるので注意．
		devno = MKDEV(major, minor);
		err = cdev_add(cdev, devno, 1);
		if ( err < 0 ) {
			int i;
			printk(KERN_INFO "%s : fail to register MG3 SCI device.\n",
					MG3_SCI_DEV_NAME);
			cdev->owner = 0;
			cdev->ops = 0;
			for ( i = 0; i < minor; ++i ) {
				cdev = &sci_dev[i].cdev;
				if ( cdev->owner != 0 ) {
					cdev_del(cdev);
					cdev->owner = 0;
					cdev->ops = 0;
				}
			}
			devno = MKDEV(major, 0);
			unregister_chrdev_region(devno, MG3_SCI_DEV_N_MAJOR);

			return err;
		}
	}


	printk(KERN_INFO "SCI driver for MG3 is registered at %d.\n", major);
	return 0; // 初期化成功
}



// モジュールの後始末．
static void sci_module_exit (void)
{
	dev_t devno;
	int minor;
	struct cdev *cdev;

	for ( minor = 0; minor < MG3_SCI_DEV_N_MINOR; ++minor ) {
		cdev = &sci_dev[minor].cdev;
		if ( cdev->owner != 0 ) {
			cdev_del(cdev);
			cdev->owner = 0;
			cdev->ops = 0;
		}
	}

	devno = MKDEV(mg3_sci_major, 0);
	unregister_chrdev_region(devno, MG3_SCI_DEV_N_MAJOR);
	printk(KERN_INFO "MG3 SCI is unregistered.\n");
}


module_init(sci_module_init);
module_exit(sci_module_exit);



// Module information
MODULE_AUTHOR("Uchibori Laboratory, Numazu National College of Technology");
MODULE_DESCRIPTION("Serial driver for MG3");
MODULE_LICENSE("GPL v2");


