/*
 *  sci.h
 *
 *	Copyright 2008-2009 
 *	Uchibori Laboratory, Numazu National College of Technology
 *	All rights reserved.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, version 2 of the License.
 */



/*
 * Base Address 0x300, 0x308
 *
 * THR/RBR　ココを直接inb/outbで読み書きすれば送受信したデータが入っている
 * アドレス：Base + 0 (R/W)
 * バッファ：1byte
 *
 *
 * IER　割込み許可フラグを与えるレジスタ。1:割込み許可　0:割込み禁止
 * アドレス：Base + 1 (R/W)
 * 0bit　受信割込み
 * 1bit　送信割込み
 * 2bit　受信ステータス割込み
 * 3bit　モデム・ステータス割込み（不使用）
 * 4bit〜7bit　"0"（何を書き込んでも"0"しか読めない。つまり不使用）
 *
 *
 * IIR　割込みID。現在、何の割込みが発生しているかを示す
 * アドレス：Base + 2 (R only)
 * 2 1 0　　←bit目を表す
 * -----
 * 0 0 1　割込み無し
 * 1 1 0　受信ステータス割込み（パリティ・エラーとか）
 * 1 0 0　受信割込み（受信完了）
 * 0 1 0　送信割込み（送信バッファが空になった。送信中だが次のデータを書込み可）
 * 0 0 0　モデム・ステータス割込み（不使用）
 * ※IIRはLSRを読むと0x01にクリアされる．
 *
 * 7〜3bitは不使用。
 *
 *
 * LCR　ライン制御・・・送受信の設定を行う
 * ※不使用
 * 0〜1bit　データ長（8bit固定予定のため不使用）
 * 2bit　ストップ・ビット長（1bit固定予定のため不使用）
 * 3bit　パリティの有無（無し固定予定のため不使用）
 * 4bit　偶数パリティか奇数パリティか。（パリティ無し固定予定のため不使用）
 * 5bit　固定パリティ（常に"1"）か？（パリティ無し固定予定のため不使用）
 * 6bit　ブレーク送信（送信データを"0"のままに保ち、相手側にブレークエラーを起こさせる）	※現在不使用
 * 7bit　分周値ラッチ（ボーレート固定予定のため不使用）
 *
 *
 * LSR　エラーとかバッファが空とかのフラグ群　0:エラーなど無し
 * アドレス：Base + 3 (R only)
 * 0bit　1:受信バッファに有効なデータ有り。（受信完了）
 * 1bit　1:オーバーランエラー（次のデータによって現在のデータが上書きされた） ※現在不使用
 * 2bit　1:パリティ・エラー（発生しない・・・筈）
 * 3bit　1:フレーミングエラー（ストップビット（本来は"1"）が正しくない）
 * 4bit　1:ブレーク受信（"0"が1byte分続いている。）
 * 5bit　1:送信バッファが空（送信中、ただし次のデータを書込み可）
 * 6bit　1:送信シフトレジスタが空（完全に送信完了）
 * 7bit　不使用
 * ※読み込みを行うと0クリアされる。（←現在コレに不具合）
 *
 */


#ifndef MG3_SCI_H
#define MG3_SCI_H

#include <linux/cdev.h>
#include <linux/spinlock.h>
#include <linux/types.h>
#include <linux/wait.h>
#include <asm/semaphore.h>
#include "sci_config.h"
#include "fifo.h"



#define MG3_SCI_THR(base_addr)  ((base_addr) + 0) // 送信バッファ (w)
#define MG3_SCI_RBR(base_addr)  ((base_addr) + 0) // 受信バッファ (r)
#define MG3_SCI_IER(base_addr)  ((base_addr) + 1) // 割り込み許可 (r/w)
#define MG3_SCI_IIR(base_addr)  ((base_addr) + 2) // 割り込み要因通知 (r)
//#define MG3_SCI_LCR(base_addr)  ((base_addr) + 3) // ライン制御 (r/w)
#define MG3_SCI_LSR(base_addr)  ((base_addr) + 3) // ラインステータス (r)



#define MG3_SCI_IER_RX 0x01 // 受信割り込み許可
#define MG3_SCI_IER_TX 0x02 // 送信割り込み許可
#define MG3_SCI_IER_ST 0x04 // 受信ステータス割り込み許可


#define MG3_SCI_IIR_MASK 0x07
#define MG3_SCI_IIR_NONE 0x01 // 割り込み無し
#define MG3_SCI_IIR_TX 0x02   // 送信割り込み
#define MG3_SCI_IIR_RX 0x04   // 受信割り込み
#define MG3_SCI_IIR_ST 0x06   // 受信ステータス割り込み


#include <motor/sci_ctl.h>
#define MG3_SCI_LSR_OVR MG3_SCIERR_OVERRUN // オーバーラン
#define MG3_SCI_LSR_PER MG3_SCIERR_PERITY  // パリティエラー
#define MG3_SCI_LSR_FER MG3_SCIERR_FEAME   // フレーミングエラー
#define MG3_SCI_LSR_BRK MG3_SCIERR_BREAK   // ブレーク受信
#define MG3_SCI_LSR_RXD 0x01 // 受信完了
#define MG3_SCI_LSR_TXR 0x20 // 送信可能
#define MG3_SCI_LSR_TXD 0x40 // 送信完了
#define MG3_SCI_LSR_BSK 0x80 // ブレーク送信





// デバッグ用のメッセージを表示する．
#define MG3_SCI_PRT_DEBUG_INFO	0x0010





// シリアルデバイスに関するあらゆる情報を納めておく構造体．
// デバイスドライバやISR（割り込み処理ルーチン）から参照される．
struct mg3_sci_dev {
	// 生じたエラーの記録
	unsigned int error;

	// read/write時にデータ(空き，完了)待ちとなったプロセスをスリープさせるための待ち行列
	wait_queue_head_t rd_queue, wr_queue, sy_queue;

	// FIFOバッファからreadできるデータ数がこの値以上になったら，rd_queueにいるスレッドを起こす
	size_t n_rd_wakeup;

	// FIFOバッファにwriteされている文字数がこの値以下になったら，wr_queueにいるスレッドを起こす
	size_t n_wr_wakeup;

	// ISRとデバイスドライバの間で同期を取る際に使用するスピンロック
	spinlock_t rd_slock, wr_slock;

	// ISRとデバイスドライバの間でデータのやりとりを行うFIFOバッファ
	struct mg3_fifo rd_buf, wr_buf;

	// スレッド間の排他制御を行うセマフォ
	struct semaphore rd_sem, wr_sem;

	// カーネルが管理するcdev構造体
	struct cdev cdev;

	// シリアルデバイスにアクセスするI/Oポートのアドレス
	const resource_size_t io_addr;

	// シリアルデバイスから発生する割り込みのIRQ
	const uint8_t irq;

	// 割込み許可フラグのコピー
	volatile uint8_t ier;

	// 割り込み要因のコピー
	volatile uint8_t iir;

	// 送受信ステータスのコピー
	volatile uint8_t lsr;

	// オプション．今のところ，デバッグ用メッセージの制御のみ．
	unsigned int opt;

	// 割り込み回数 デバッグに使っているんだけど，OSに標準であったりして．
	unsigned int n_intr;

	// デバイスのopen状況を記録するためのマジックナンバー
	uint32_t magic;
};


typedef struct mg3_sci_dev SciDevice;





#endif // MG3_SCI_H
