/*
 *
 * ioctl.c
 *	test ioctl()
 *
 *
 * Copyright 2008-2009 
 * Uchibori Laboratory, Numazu National College of Technology
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <termios.h>
#include <motor/sci_ctl.h>
#include "devfile.h"


#define BUF_SIZE (MG3_SCI_FIFO_SIZE  + 1)



static int rw_n (int fd, void *buf, size_t n)
{
	size_t i;
	unsigned char *wbuf = buf;

	for ( i = 0; i < n; ++i ) {
		wbuf[i] = i;
	}

	errno = 0;
	return write(fd, buf, n);
}


static void overrun (int fd)
{
	int err, line_state;
	char buf[BUF_SIZE];


	fprintf(stderr, "  checking over run...");
	err = rw_n(fd, buf, BUF_SIZE);
	assert(err > 0 && err <= BUF_SIZE);
	usleep(100000);
	err = read(fd, buf, BUF_SIZE);
	assert(err > 0 && err <= BUF_SIZE);
	err = read(fd, buf, BUF_SIZE);
	assert(err < 0 && errno == EIO);
	line_state = ioctl(fd, MG3_SCI_IOCGLSR);
	assert(line_state == MG3_SCIERR_OVERRUN);
	fprintf(stderr, "ok.\n");
}


static void reseterr (int fd)
{
	int err, line_state;
	char buf[BUF_SIZE];


	fprintf(stderr, "  checking clear error...");
	err = rw_n(fd, buf, BUF_SIZE); // over runさせる
	usleep(100000);

	line_state = ioctl(fd, MG3_SCI_IOCGLSR);
	assert(line_state == MG3_SCIERR_OVERRUN); // over runしているのを確認

	err = ioctl(fd, MG3_SCI_IOCCLSR); // エラーをクリア
	assert(err == 0);
	line_state = ioctl(fd, MG3_SCI_IOCGLSR);
	assert(line_state == 0); // クリアされているか確認


	fprintf(stderr, "ok.\n");
}


#define IOSIZE 4
static void bufproc (int fd)
{
	int err, n;
	char buf[BUF_SIZE];


	// TCFLSH(TCIOFLUSH)
	fprintf(stderr, "  checking control FIFO...");
	err = ioctl(fd, TCFLSH, TCIOFLUSH);
	assert(err == 0);
	err = ioctl(fd, TIOCINQ, &n);
	assert(err == 0);
	assert(n == 0);
	err = ioctl(fd, TIOCOUTQ, &n);
	assert(err == 0);
	assert(n == 0);

	// 出力バイト数 TIOCOUTQ
	err = write(fd, buf, IOSIZE);
	assert(err == IOSIZE);
	err = ioctl(fd, TIOCOUTQ, &n);
	assert(err == 0);
	assert(n > 0);
	usleep(100000);

	// 入力バイト数 TIOCINQ
	err = ioctl(fd, TIOCINQ, &n);
	assert(err == 0);
	assert(n == IOSIZE);
	err = read(fd, buf, IOSIZE);
	assert(err == IOSIZE);

	// 出力バッファクリア TCFLSH(TCOFLUSH)
	err = write(fd, buf, IOSIZE);
	assert(err == IOSIZE);
	err = ioctl(fd, TIOCOUTQ, &n);
	assert(err == 0);
	assert(n > 0);
	err = ioctl(fd, TCFLSH, TCOFLUSH);
	assert(err == 0);
	err = ioctl(fd, TIOCOUTQ, &n);
	assert(err == 0);
	assert(n == 0);
	usleep(100000);

	// 入力バッファクリア TCFLSH(TCIFLUSH)
	syslog(LOG_INFO, "flush input");
	err = ioctl(fd, TIOCINQ, &n);
	assert(err == 0);
	assert(n > 0);
	err = ioctl(fd, TCFLSH, TCIFLUSH);
	assert(err == 0);
	err = ioctl(fd, TIOCINQ, &n);
	assert(err == 0);
	assert(n == 0);
	fprintf(stderr, "ok.\n");
}



void ck_break (int fd)
{
	int err;
	char c;

	// ブレーク
	fprintf(stderr, "  sending break...\a");
	ioctl(fd, TIOCSBRK);
	usleep(3000000);
	ioctl(fd, TIOCCBRK);
	usleep(100000);
	/*
	err = read(fd, &c, sizeof(c));
	assert(err < 0);
	assert(errno == EIO);
	err = ioctl(fd, MG3_SCI_IOCGLSR);
	assert(err == MG3_SCIERR_FEAME); // break信号を受信しているか確認
	*/

	fprintf(stderr, "done.\a\n");
}



int main (void)
{
	int fd;


	fprintf(stderr, "testing ioctl...\n");
	fd = open(FNAME, O_RDWR);
	if ( fd < 0 ) {
		perror("open");
		return -1;
	}



	overrun(fd);
	reseterr(fd);
	bufproc(fd);
	usleep(1000000);
	ioctl(fd, MG3_SCI_IOENDMSG);
	ck_break(fd);


	close(fd);
	
	fprintf(stderr, "done.\n");

	return 0;
}

