/*
 *
 * mt.c
 *	Stress test in multi-thread
 *
 *
 * Copyright 2008-2009 
 * Uchibori Laboratory, Numazu National College of Technology
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */


#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <motor/sci_ctl.h>
#include "devfile.h"

#define RD_DEV_FILE FNAME
#define WR_DEV_FILE FNAME
#define N_RW 10
#define N_THREAD 100
#define TEST_STRING "mirs"


int rd_fd, wr_fd;

pthread_mutex_t rd_time_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t wr_time_lock = PTHREAD_MUTEX_INITIALIZER;
volatile int rd_time = 0;
volatile int wr_time = 0;


pthread_t wr_thread[N_THREAD];
pthread_t rd_thread[N_THREAD];

pthread_mutex_t exec_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t  exec_cond = PTHREAD_COND_INITIALIZER;
int ready_to_exec = 0;



#define BUF_LEN 128
static void *rd_func (void *arg)
{
	int i;
	int err;
	size_t size = strlen(TEST_STRING) + 1;
	char buf[BUF_LEN];

	assert(size <= BUF_LEN);
	arg = NULL; // fake gcc


	//
	// 実行を待ち合わせる．
	//
	err = pthread_mutex_lock(&exec_lock);
	if ( err != 0 ) {
		errno = err;
		perror("Lock mutex");
		exit(-1);
	}
	while ( ready_to_exec == 0 ) {
		err = pthread_cond_wait(&exec_cond, &exec_lock);
		if ( err != 0 ) {
			errno = err;
			perror("Cond wait");
			exit(-1);
		}
	}
	err = pthread_mutex_unlock(&exec_lock);
	if ( err != 0 ) {
		errno = err;
		perror("Unlock mutex");
		exit(-1);
	}


	//
	// 読み込み
	//
	for ( i = 0; i < N_RW; i++ ) {
		err = read(rd_fd, buf, size);
		if ( err < 0 || (size_t)err != size ) {
			perror("Write");
			exit(-1);
		}
		assert(strcmp(buf, TEST_STRING) == 0);

		err = pthread_mutex_lock(&rd_time_lock);
		if ( err != 0 ) {
			errno = err;
			perror("Lock mutex");
			exit(-1);
		}
		rd_time += 1;
		err = pthread_mutex_unlock(&rd_time_lock);
		if ( err != 0 ) {
			errno = err;
			perror("Unlock mutex");
			exit(-1);
		}
	}

	return 0;
}


static void *wr_func (void *arg)
{
	int i;
	int err;
	size_t size = strlen(TEST_STRING) + 1;

	arg = NULL; // fake gcc

	// detach
	pthread_detach(pthread_self());


	//
	// 実行を待ち合わせる．
	//
	err = pthread_mutex_lock(&exec_lock);
	if ( err != 0 ) {
		errno = err;
		perror("Lock mutex");
		exit(-1);
	}
	while ( ready_to_exec == 0 ) {
		err = pthread_cond_wait(&exec_cond, &exec_lock);
		if ( err != 0 ) {
			errno = err;
			perror("Cond wait");
			exit(-1);
		}
	}
	err = pthread_mutex_unlock(&exec_lock);
	if ( err != 0 ) {
		errno = err;
		perror("Unlock mutex");
		exit(-1);
	}


	//
	// 書き込み
	//
	for ( i = 0; i < N_RW; i++ ) {
		err = write(wr_fd, TEST_STRING, size);
		if ( err < 0 || (size_t)err != size ) {
			perror("Write");
			exit(-1);
		}

		err = pthread_mutex_lock(&wr_time_lock);
		if ( err != 0 ) {
			errno = err;
			perror("Lock mutex");
			exit(-1);
		}
		wr_time += 1;
		err = pthread_mutex_unlock(&wr_time_lock);
		if ( err != 0 ) {
			errno = err;
			perror("Unlock mutex");
			exit(-1);
		}

	}

	return 0;
}




int create_thread (void)
{
	int err;
	int policy;
	int i;
	pthread_attr_t thread_attr;
	struct sched_param schd_prm;


	//
	// 書き込みスレッド用attributeを初期化する．
	//
	err = pthread_attr_init(&thread_attr);
	if ( err != 0 ) {
		return err;
	}

	err = pthread_attr_setinheritsched(&thread_attr, PTHREAD_EXPLICIT_SCHED);
	if ( err != 0 ) {
		return err;
	}

	err = pthread_attr_setschedpolicy(&thread_attr, SCHED_RR);
	if ( err != 0 ) {
		return err;
	}

	err = pthread_getschedparam(pthread_self(), &policy, &schd_prm);
	if ( err != 0 ) {
		return err;
	}

	schd_prm.sched_priority = 50;
	err = pthread_attr_setschedparam(&thread_attr, &schd_prm);
	if ( err != 0 ) {
		return err;
	}


	//
	// 書き込みスレッドを作成する．
	//
	for ( i = 0; i < N_THREAD; i++ ) {
		err = pthread_create(&wr_thread[i], &thread_attr,
			wr_func, NULL);
		if ( err != 0 ) {
			return err;
		}
	}


	//
	// 読み込みスレッド用attributeを初期化する．
	//
	schd_prm.sched_priority += 10;
	err = pthread_attr_setschedparam(&thread_attr, &schd_prm);
	if ( err != 0 ) {
		return err;
	}


	//
	// 読み込みスレッドを作成する．
	//
	for ( i = 0; i < N_THREAD; i++ ) {
		err = pthread_create(&rd_thread[i], &thread_attr,
			rd_func, NULL);
		if ( err != 0 ) {
			return err;
		}
	}


	return 0;
}




int main (void)
{
	int err;
	int i;


	if ( strcmp(RD_DEV_FILE, WR_DEV_FILE) == 0 ) {
		wr_fd = open(WR_DEV_FILE, O_RDWR);
		if ( wr_fd < 0 ) {
			perror("Open");
			return -1;
		}
		rd_fd = wr_fd;
		ioctl(rd_fd, MG3_SCI_IODSDMSG);
	}
	else {
		fprintf(stderr, "not implement yet\n");
		return -1;
	}

	err = create_thread();
	if ( err != 0 ) {
		errno = err;
		perror("Create thread");
		return -1;
	}


	// Start!!
	fprintf(stderr, "stress test.");
	err = pthread_mutex_lock(&exec_lock);
	if ( err != 0 ) {
		errno = err;
		perror("Lock mutex");
		exit(-1);
	}
	ready_to_exec = 1;
	err = pthread_cond_broadcast(&exec_cond);
	if ( err != 0 ) {
		errno = err;
		perror("Cond broadcast");
		exit(-1);
	}
	err = pthread_mutex_unlock(&exec_lock);
	if ( err != 0 ) {
		errno = err;
		perror("Unlock mutex");
		exit(-1);
	}
	fprintf(stderr, ".");



	// wait
	for ( i = 0; i < N_THREAD; i++ ) {
		void *ret_val;
		err = pthread_join(rd_thread[i], &ret_val);
		if ( err != 0 ) {
			errno = err;
			perror("Join");
			return -1;
		}
		assert(ret_val == 0);
	}
	fprintf(stderr, ".");

	assert(rd_time == N_RW * N_THREAD);
	assert(wr_time == N_RW * N_THREAD);

	fprintf(stderr, "ok.\n");

	return 0;
}




